#include <stdio.h>

#define NEWPRIM(cname)  void cname(struct forth_instance *fi)
#define REGPRIM(cname, fname)  reg_prim(fi, fname, &cname)
#define push(x)  st_push(fi, (int) x)
#define pop()  st_pop(fi);


NEWPRIM(read_func) {
  int fd, len;
  char *dest;

  len = pop();
  dest = (char *) pop();
  fd = pop();

  push(read(fd, dest, len));
}

NEWPRIM(poll_func) {
  /* FIXME: Some sort of API for polling multiple fds... */
  /* Hell, make the poll API actually useful. :) */
  struct pollfd tp;
  int tv;

  tv = pop();
  tp.fd = pop();
  tp.events = POLLIN;

  push(poll(&tp, 1, tv));
}

NEWPRIM(term_raw_on) {
  struct termios tptty;

  ioctl(0, IOCTL_GET, &tptty);
  tptty.c_lflag &= ~(ICANON | ECHO);
  if (ioctl(0, IOCTL_SET, &tptty) == -1) {
    if(fi->output) fprintf(fi->output, "IOCTL FAILURE\n");
    error(fi);
  }
}

NEWPRIM(term_raw_off) {
  struct termios tptty;

  ioctl(0, IOCTL_GET, &tptty);
  tptty.c_lflag |= (ICANON | ECHO);
  if (ioctl(0, IOCTL_SET, &tptty) == -1) {
    if(fi->output) fprintf(fi->output, "IOCTL FAILURE\n");
    error(fi);
  }
}

NEWPRIM(flush) {
  fflush(stdin);
  fflush(stdout);
}


NEWPRIM(microsleep) {
  int tp;
  tp = pop();
  if (tp>=999999) for(;tp>=999999;tp-=999999) usleep(999999);
  usleep(tp);
}


NEWPRIM(time_fetch) {
  push(time(NULL));
}


NEWPRIM(gettime) {
  struct timeval tp;
  gettimeofday(&tp, NULL);
  push(tp.tv_usec);
  push(tp.tv_sec);
}


void external_prims(struct forth_instance *fi) {

  REGPRIM(read_func, "read");
  REGPRIM(poll_func, "poll");
  REGPRIM(term_raw_on, "term-raw-on");
  REGPRIM(term_raw_off, "term-raw-off");
  REGPRIM(flush, "flush");
  REGPRIM(microsleep, "usleep");
  REGPRIM(time_fetch, "time@");
  REGPRIM(gettime, "gettime");

}
