: \ 10 parse drop drop ; immediate
: ( 41 parse drop drop ; immediate
\ The above 2 definitions are for comments. DO NOT REMOVE!

\ HardCore SoftWare's FRUGAL system

\ These are the FORTH words that aren't defined in the core, but they
\ are essential language elements.

\ This program is Copyright 2002 by HardCore SoftWare.
\ See the file COPYING for details, or visit www.gnu.org


\ ---------------------- FUNDAMENTALS
: dup 0 pick ;
: over 1 pick ;
: swap 1 roll ;
: rot 2 roll ;
: nip swap drop ;
: tuck dup swap rot ;
: 2swap 3 3 roll roll ;
: 2drop drop drop ;
: 2over 3 pick 3 pick ;
: 2dup 1 pick 1 pick ;
: 0= 0 = ;
: <> = 0= ;
: <= > 0= ;
: >= < 0= ;
: . printnum 32 emit ;
: i r> r> dup rot >r >r ;
: j r> r> r> r> dup 4 roll >r >r >r >r ;



\ ---------------------- TIME STUFF, ETC
: ms 1000 * usleep ;


\ ---------------------- HEAP STUFF, ETC
: here h @ ;
: allot h @ + h ! ;
: constant create 2 ( Hardwired VM_NUMBER ) , , compile-exit ;
: variable create here 16 + dup 2 ( Hardwired VM_NUMBER ) , ,
           compile-exit 0 ,  ! ;
: const constant ;
: var variable ;
: 2variable variable ;
: cell+ 4 + ;
: cells 4 * ;
: char+ 1 + ;
: chars ;
: 2! swap over ! cell+ ! ;
: 2@ dup cell+ @ swap @ ;



\ ---------------------- SYSTEM CONSTANTS
0 constant STDIN
1 constant STDOUT

1 constant IMMEDIATE
2 constant COMPILE-ONLY
4 constant UNSEEABLE

1 constant VM_PRIMITIVE
2 constant VM_NUMBER
3 constant VM_BRANCH
4 constant VM_IFBRANCH

10 constant HEADER_LEN

0 constant 0
1 constant 1


\ ---------------------- COMPILING WORDS
\ Thanks to onetom on #forth for these next 3:
: #,  VM_NUMBER , ;
: branch  VM_BRANCH , ; \ Changed slightly by me
: ?branch  VM_IFBRANCH , ; \ Changed slightly by me
\ compile is a non-traditional forth word.
\ It's another approach to the same thing does> does. :)
: compile ' #, , ; immediate
: postpone ' , ; immediate compile-only
: [ 0 state ! ;
: ] 1 state ! ;
: exit compile-exit ; compile-only immediate
: unseeable l @ cell+ @ UNSEEABLE or l @ cell+ ! ;
: isunseeable HEADER_LEN cells - cell+ @ UNSEEABLE and ;




\ ---------------------- CONTROL FLOW WORDS
: if compile 0= , ?branch here 0 , ; immediate compile-only
: else branch here 0 , swap here swap ! ; immediate compile-only
: then here swap ! ; immediate compile-only

: begin here ; immediate compile-only
: again branch , ; immediate compile-only
: while ?branch , ; immediate compile-only
: until compile 0= , ?branch , ; immediate compile-only
: do compile swap , compile >r , compile >r , here ; immediate compile-only
: +loop
        compile r> , compile + , compile dup , compile i , compile <> ,
        compile swap , compile >r ,
        ?branch , 
        compile r> , compile drop , compile r> , compile drop ,
        ; immediate compile-only
: loop #, 1 , postpone +loop ; immediate compile-only



\ ---------------------- MATH, ETC
: negate -1 * ;
: 1+ 1 + ;
: 1- 1 - ;
: 2* 2 * ;
: 2/ 2 / ;
: min 2dup < if drop else nip then ;
: max 2dup > if drop else nip then ;
: abs dup 0 < if -1 * then ;



\ ---------------------- MEMORY, DIRECT STACK MANIPULATION, ETC
: +! dup @ rot rot + swap ! ;
: ? @ . ;
: depth sp s0 - 4 / ;
  ( These next 2 are right out of the ANSI standard. )
: movecell ( addr1-src addr2-dest -- ) swap @ swap ! ;
: move 0 do 2dup i cells + swap i cells + swap movecell loop 2drop ;


\ FIXME: type should be ( blah bleh -- ), not ( blah bleh -- blurg )
\ ---------------------- I/O, ETC
: cr 10 emit ;
: bl 32 ;
: space bl emit ;
: spaces dup 0 > if 0 do space loop else drop then ;
: type 0 do dup c@ emit char+ loop ;
: char 32 parse drop c@ ;
: char, char #, , ; immediate
: movechar ( caddr1-src caddr2-dest -- ) swap c@ swap c! ;
: cmove 0 do 2dup i chars + swap i chars + swap movechar loop 2drop ;
: s" here 34 parse dup rot here swap dup allot cmove ; compile-only
: ." unseeable #, here >r 0 , #, here >r 0 ,
    compile type , branch here 0 ,
    s" r> ! r> ! here swap ! compile drop , compile flush ,
   ; immediate compile-only

: accept STDIN rot read ;  
: key term-raw-on
      0 ( This 0 will contain the char )
      sp 1 cells -  STDIN swap 1 read drop
      term-raw-off ;
: rawkey 0 ( This 0 will contain the char )
         sp 1 cells -  STDIN swap 1 read drop ;



\ ---------------------- TERM
: emchar #, char , compile emit , ; immediate
: emesc 27 emit ;

\ COLOURS
0 constant black
1 constant red
2 constant green
3 constant yellow
4 constant blue
5 constant magenta
6 constant cyan
7 constant white

\ ATTRIBUTES
0 constant restore
1 constant bright
2 constant dim
4 constant underscore
5 constant blink
7 constant reverse
8 constant hidden

: term-fg ( colour -- ) 30 + emesc emchar [ printnum emchar m ;
: term-bg ( colour -- ) 40 + emesc emchar [ printnum emchar m ;
: term-attr ( attr -- ) emesc emchar [ printnum emchar m ;
: term-cls emesc emchar [ emchar 2 emchar J ;
: term-wrap-on emesc emchar [ emchar 7 emchar h ;
: term-wrap-off emesc emchar [ emchar 7 emchar l ;
: term-cur-on emesc emchar [ emchar ? 25 printnum emchar h ;
: term-cur-off emesc emchar [ emchar ? 25 printnum emchar l ;
: term-home emesc emchar [ emchar H ;
: term-xy ( x y -- ) emesc emchar [ printnum emchar ; printnum emchar H ;
: term-up emesc emchar [ emchar A ;
: term-down emesc emchar [ emchar B ;
: term-forward emesc emchar [ emchar C ;
: term-backward emesc emchar [ emchar D ;
: term-save emesc emchar 7 ;
: term-unsave emesc emchar 8 ;
: term-enable-scroll ( start end -- ) 
       swap emesc emchar [ printnum emchar ; printnum emchar r ;
: term-scroll-down emesc emchar [ emchar D ;
: term-scroll-up emesc emchar [ emchar M ;
: term-clear-to-eol emesc emchar [ emchar K ;
: term-clear-to-sol emesc emchar [ emchar 1 emchar K ;
: term-clear-line emesc emchar [ emchar 2 emchar L ;
: term-clear-to-top emesc emchar [ emchar J ;
: term-clear-to-bot emesc emchar [ emchar 1 emchar J ;





\ ---------------------- TOOLS, ETC
\ : abort" ?branch here 0 , compile-exit here swap !
\  #, compile abort , compile >r , postpone ." ; immediate compile-only
: .s 60 emit space depth . 62 emit space depth 1 < if exit then 
    depth 0 do s0 i cells + ? loop ." [TOP]" ;
: forget ' HEADER_LEN cells - dup @ l ! h ! ;
\ print-word Takes an address from the start of the header.
\ 'print from the beginning of the actual code (where ' returns)
: print-word ( addr -- ) cell+ cell+ dup char+ swap c@ type drop ;
: 'print HEADER_LEN cells - print-word ;
: bytes-and-words ( -- bytes words )
  0 l begin @ swap 1+ swap dup @ 0= until
  here swap -   swap ;
: words ." ( " bytes-and-words . ." WORDS  - " . ." BYTES )" cr l @
  begin dup print-word space @ dup 0= until drop ;
: procprint ( addr -- nextaddr )
  dup . ." : "
  dup @ VM_NUMBER = if cell+ dup @ ." LITERAL: " . cr cell+ exit then
  dup @ VM_BRANCH = if cell+ dup @ ." VM_BRANCH -> " . cr cell+ exit then
  dup @ VM_IFBRANCH = if cell+ dup @ ." VM_IFBRANCH -> " . cr cell+ exit then
  dup @ 'print cr cell+ ;
: see ' dup isunseeable if ." UNSEEABLE" cr drop exit then 
  dup @ VM_PRIMITIVE = if ." PRIMITIVE" cr drop exit then
  begin procprint dup @ VM_PRIMITIVE = until . ." : EXIT" cr ;
: abort ." RESETING STACKS AND STATE" cr quit ;



\ ---------------------- DEBUGGING
( These next 4 from onetom on #forth. If x is a variable use "x dump-line" )
: h. dup c@ . char+ ;
: c. dup c@ dup 0= if drop 46 then emit char+ ;
: 8h. h. h. h. h. h. h. h. h. ; : 8c. c. c. c. c. c. c. c. c. ;
: dump-line 8h. ." - " 8h. 16 - 3 spaces 8c. ."  - " 8c. drop ;


\ ---------------------- RANDOM
gettime nip var rand-num
: rand rand-num @ 31421 * 6927 + dup rand-num ! ;
: choose dup 2 < if ." CHOOSE DOES NOT ACCEPT NUMBERS BELOW 2" cr abort then
         rand swap mod abs ;


\ ---------------------- INDIRECTION
: icreate create compile exit , compile-exit ; 
: iset> ' swap ! ;


\ ---------------------- WRAP UP
: bye term-raw-off die ;
: intro ." HardCore SoftWare's FORTH system: FRUGAL   "
  ver char, V emit dup dup 100 / printnum
    char, . emit  100 mod 10 / printnum  char, . emit  10 mod printnum
  cr cr ;
intro
